from abc import ABC, abstractmethod

import numpy as np

from dtcontrol.dataset.scots_dataset_loader import ScotsDatasetLoader
from dtcontrol.dataset.uppaal_dataset_loader import UppaalDatasetLoader
from dtcontrol.dataset.vector_dataset_loader import VectorDatasetLoader
from dtcontrol.dataset.csv_dataset_loader import CSVDatasetLoader
from dtcontrol.util import get_filename_and_ext


class Dataset(ABC):
    """
    Attributes:
        X_train: the training data of shape (num_states, num_state_dims)
        Y_train:
            Multi-output: the training labels of shape (num_input_dims, num_states, max_non_determinism)
                For example:  [ [[1  2 -1]  [[4  5 -1]
                                [2  3 -1]   [4  6 -1]
                                [1  2  3]   [4  6  6]
                                [1 -1 -1]]  [5 -1 -1]] ]
                has the shape (2, 4, 3) as there are two control inputs, 4 states and
                at most 3 non-deterministic choices for each state.

                The Y_train in the above example can be thought of a list [y1, y2]
                where y1 gives the values for the first control input and y2 gives
                the values for the second control input. Let us for example see
                which inputs are allowed for the 2nd state.

                y1[1] = [2  3 -1] and y2[1] = [4  6 -1]

                This means that the allowed control inputs for the 2nd state are
                (2, 4) and (3, 6). The -1 is a filler just to make the length of
                the lists  = max_non_determinism.

                Use np.stack(Y_train, axis=2) in order to get an array of the form
                [[[1  4], [ 2  5], [-1 -1]],
                 [[2  4], [ 3  6], [-1 -1]]
                 [[1  4], [ 2  6], [ 3  6]]
                 [[1  5], [-1 -1], [-1 -1]]

                from which it is easier to extract control actions as tuples.
            Single-output: the training labels of shape (num_states, max_non_determinism)
                For example:   [[1  2 -1]
                                [2  3 -1]
                                [1  2  3]
                                [1 -1 -1]]
                has the shape (4, 3) as there are 4 states and at most 3 non-deterministic
                choices for each state.

                Y_train[i] gives the allowed control inputs for the ith state. -1 is a
                filler just to make the length of the lists  = max_non_determinism.
    """
    def __init__(self, filename):
        self.filename = filename
        self.name, self.extension = get_filename_and_ext(filename)
        self.extension_to_loader = {
            '.vector': VectorDatasetLoader(),
            '.scs': ScotsDatasetLoader(),
            '.dump': UppaalDatasetLoader(),
            '.csv': CSVDatasetLoader(),
        }
        if self.extension not in self.extension_to_loader:
            raise ValueError('Unknown file format.')
        self.X_train = None
        self.X_metadata = {"variables": None, "min": None, "max": None, "step_size": None}
        self.Y_train = None
        self.Y_metadata = {"variables": None, "min": None, "max": None, "step_size": None, 'num_rows': None, 'num_flattened': None}
        self.index_to_value = {}
        self.is_deterministic = None

    def load_if_necessary(self):
        if self.X_train is None:
            self.X_train, self.X_metadata, self.Y_train, self.Y_metadata, self.index_to_value = self.extension_to_loader[
                self.extension].load_dataset(self.filename)
            self.Y_metadata['num_rows'] = len(self.X_train)
            self.Y_metadata['num_flattened'] = sum(1 for row in self.Y_train for y in row)

    def load_metadata_from_json(self, json):
        metadata = json['metadata']
        self.X_metadata = metadata['X_metadata']
        self.Y_metadata = metadata['Y_metadata']

    def check_loaded(self):
        if self.X_train is None:
            raise RuntimeError('Dataset is not loaded.')

    @abstractmethod
    def compute_accuracy(self, Y_pred):
        pass

    @abstractmethod
    def get_unique_labels(self):
        pass

    @abstractmethod
    def map_unique_label_back(self, label):
        pass

    """
    Computes unique labels of a 2d label array by mapping every unique inner array to an int. Returns the unique labels
    and the int mapping.
    """

    @staticmethod
    def _get_unique_labels(labels):
        l = []
        int_to_label = {}
        next_unused_int = 1  # OC1 expects labels starting with 1
        label_str_to_int = {}
        for i in range(len(labels)):
            label_str = ','.join(sorted([str(i) for i in labels[i] if i != -1]))
            if label_str not in label_str_to_int:
                label_str_to_int[label_str] = next_unused_int
                int_to_label[next_unused_int] = labels[i]
                next_unused_int += 1
            new_label = label_str_to_int[label_str]
            l.append(new_label)
        return np.array(l), int_to_label

    @staticmethod
    def _get_max_labels(labels):
        flattened_labels = labels.flatten()
        # remove -1 as we use it only as a filler
        flattened_labels = flattened_labels[flattened_labels != -1]
        label_counts = np.bincount(flattened_labels)
        new_labels = []
        for i in range(len(labels)):
            current = labels[i]
            current = current[current != -1]
            max_label = max(list(current), key=lambda l: label_counts[l])
            assert max_label != -1
            new_labels.append(max_label)
        return np.array(new_labels)



from sklearn.linear_model import LogisticRegression

from dtcontrol.benchmark_suite import BenchmarkSuite
from dtcontrol.classifiers.linear_classifier_dt import LinearClassifierDT
from dtcontrol.classifiers.oc1_wrapper import OC1Wrapper
from dtcontrol.classifiers.cart_custom_dt import CartDT
from dtcontrol.classifiers.max_freq_dt import MaxFreqDT
from dtcontrol.classifiers.bdd import BDD

suite = BenchmarkSuite(timeout=60 * 60 * 2, save_folder='saved_classifiers', benchmark_file='benchmark_tmp', rerun=True)
suite.add_datasets(['examples'],
                   include=["cartpole",
                            # "tworooms-noneuler-latest",
                            # "helicopter",
                            # "cruise-latest",
                            # "dcdc",
                            # "10rooms",
                            # "truck_trailer",
                            # "traffic_1m",
                            # "traffic_10m",
                            # "traffic_30m",
                            # "vehicle",
                            # "aircraft"
                            ],
                   exclude=[
                       'aircraft', 'traffic_30m', 'truck_trailer'
                   ]
                   )
classifiers = [
    CartDT(),
    # LinearClassifierDT(LogisticRegression, solver='lbfgs', penalty='none'),
    # LinearClassifierDecisionTree(LinearSVC, max_iter=5000),
    # MaxFreqDT(),
    # MaxLCDecisionTree(LogisticRegression, solver='lbfgs', penalty='none'),
    # OC1Wrapper(num_restarts=1, num_jumps=1),
    # BDD()
    # MaxEveryNodeMultiDecisionTree()
]
suite.benchmark(classifiers)



#!/usr/bin/env python

"""
README

Run dtcontrol --help to see usage. Some examples are given below.

Example:
dtcontrol --input controller.scs --output decision_trees --method maxcart

will read controller.scs and use maxcart on it and print the c, dot, vhdl files
into the decision_trees folder


dtcontrol --input controller1.scs controller2.scs --output decision_trees --benchmark_file benchmarks.json

will read controller1.scs and controller2.scs, try out all methods and save the
results in decision_trees; moreover save the run and tree statistics in
benchmark.json and a nice HTML table in benchmark.html


dtcontrol --input dumps --output decision_trees --method all --determinize maxfreq minnorm

will read all valid controllers in dumps, run the determinized variants (using the maxfreq
 and minnorm determinization strategies) of all methods on them and save the decision
trees in decision_trees

"""

import argparse
import logging
import re
import sys
from os import makedirs
from os.path import exists, isfile, splitext

import pkg_resources
from sklearn.linear_model import LogisticRegression
from sklearn.svm import LinearSVC

from dtcontrol.benchmark_suite import BenchmarkSuite
from dtcontrol.classifiers.cart_custom_dt import CartDT
from dtcontrol.classifiers.linear_classifier_dt import LinearClassifierDT
from dtcontrol.classifiers.max_freq_dt import MaxFreqDT
from dtcontrol.classifiers.max_freq_linear_classifier_dt import MaxFreqLinearClassifierDT
from dtcontrol.classifiers.max_freq_multi_dt import MaxFreqMultiDT
from dtcontrol.classifiers.norm_dt import NormDT
from dtcontrol.classifiers.norm_linear_classifier_dt import NormLinearClassifierDT
from dtcontrol.classifiers.oc1_wrapper import OC1Wrapper
from dtcontrol.classifiers.random_dt import RandomDT


def main():
    def is_valid_file_or_folder(parser, arg):
        if not exists(arg):
            parser.error(f"The file/folder {arg} does not exist")
        else:
            return arg

    def is_valid_file(parser, arg):
        if not isfile(arg):
            parser.error(f"The file {arg} does not exist. Give a valid JSON file path.")
        else:
            return arg

    def parse_timeout(timeout_string: str):
        """
        Parses the timeout string

        :param timeout_string: string describing timeout - an integer suffixed with s, m or h
        :return: timeout in seconds
        """
        # Default timeout set to 2 hours
        unit_to_factor = {'s': 1, 'm': 60, 'h': 3600}
        if re.match(r'^[0-9]+[smh]$', timeout_string):
            factor = unit_to_factor[timeout_string[-1]]
            timeout = int(args.timeout[:-1]) * factor
        else:
            # In case s, m or h is missing; then interpret number as timeout in seconds
            try:
                timeout = int(timeout_string)
            except ValueError:
                parser.error("Invalid value passed as timeout.")
        return timeout

    def get_classifiers(methods, det_strategies):
        """
        Creates classifier objects for each method

        :param methods: list of method strings
        :param det_strategies: list of determinization strategies
        :return: list of classifier objects
        """
        method_map = {
            'cart': {
                'none': [CartDT()],  # TODO: remove lists and directly map to classifiers
                'maxfreq': [MaxFreqDT()],
                'minnorm': [NormDT(min)],
                # 'random': [RandomDT()],
            },
            'linsvm': {
                'none': [LinearClassifierDT(LinearSVC, max_iter=5000)],
                'maxfreq': [MaxFreqLinearClassifierDT(LinearSVC, max_iter=5000)],
                'minnorm': [NormLinearClassifierDT(min, LinearSVC, max_iter=5000)],
            },
            'logreg': {
                'none': [LinearClassifierDT(LogisticRegression, solver='lbfgs', penalty='none')],
                'maxfreq': [MaxFreqLinearClassifierDT(LogisticRegression, solver='lbfgs', penalty='none')],
                'minnorm': [NormLinearClassifierDT(min, LogisticRegression, solver='lbfgs', penalty='none')],
            },
            'oc1': {
                'none': [OC1Wrapper(num_restarts=20, num_jumps=5)]
            }
        }

        # construct all possible method - determinization strategy combinations
        classifiers = []

        if 'all' in methods:
            methods = method_map.keys()

        for method in methods:
            if method not in method_map:
                logging.warning(f"No method '{method}' exists. Skipping...")
                continue

            if 'all' in det_strategies:
                classifiers.extend(
                    [classifier for cls_group in method_map[method].values() for classifier in cls_group])
            else:
                for det_strategy in det_strategies:
                    if det_strategy not in method_map[method]:
                        logging.warning(f"Method '{method}' and determinization strategy '{det_strategy}' "
                                        f"don't work together (yet). Skipping...")
                        continue
                    classifiers.extend(method_map[method][det_strategy])

        # returns a flattened list
        return classifiers

    logging.basicConfig(level=logging.INFO, format='%(message)s')

    parser = argparse.ArgumentParser(prog="dtcontrol")

    version = pkg_resources.require("dtcontrol")[0].version
    parser.add_argument("-v", "--version", action='version',
                        version=f'%(prog)s {version}')

    parser.add_argument("--input", "-i", nargs="+", type=(lambda x: is_valid_file_or_folder(parser, x)),
                        help="The input switch takes in one or more space separated file names or "
                             "a folder name which contains valid controllers (.scs, .dump or .csv)")

    parser.add_argument("--method", "-m", default=['all'], nargs="+",
                        help="The method switch takes in one or more space separated method names as "
                             "arguments. Available methods are: 'cart', 'linsvm', 'logreg', 'oc1'. Running "
                             "with --method 'all' will run all possible methods. For description about each method, "
                             "see manual. If this switch is omitted, defaults to 'all'")

    parser.add_argument("--determinize", "-d", nargs='+', metavar='DETSTRATEGY', default=['none'],
                        help="In case of non-deterministic controllers, specify, if desired, the determinization "
                             "strategy. Possible options are 'minnorm' and 'maxfreq'. If the option 'none' is passed, "
                             "then the controller is not determinized. The shorthand '-d all' is equivalent to "
                             "'-d none maxfreq minnorm'.")

    parser.add_argument("--timeout", "-t", type=str,
                        help="Sets a timeout for each method. Can be specified in seconds, minutes "
                             "or hours (eg. 300s, 7m or 3h)")

    parser.add_argument("--benchmark-file", "-b", metavar="FILENAME", type=str,
                        help="Saves statistics pertaining the construction of the decision trees and their "
                             "sizes into a JSON file, and additionally allows to view it via an HTML file.")

    parser.add_argument("--output", "-o", type=str,
                        help="The output switch takes in a path to a folder where the constructed controller "
                             "representation would be saved (c and dot)")

    parser.add_argument("--rerun", "-r", action='store_true',
                        help="Rerun the experiment for all input-method combinations. Overrides the default "
                             "behaviour of not running benchmarks for combinations which are already present"
                             " in the benchmark file.")

    args = parser.parse_args()

    kwargs = dict()

    if args.input:
        dataset = args.input
    else:
        parser.print_help()
        sys.exit()

    kwargs["timeout"] = 2 * 60 * 60
    if args.timeout:
        kwargs["timeout"] = parse_timeout(args.timeout)

    if args.benchmark_file:
        filename, file_extension = splitext(args.benchmark_file)
        kwargs["benchmark_file"] = filename
    else:
        kwargs["benchmark_file"] = 'benchmark'  # TODO best practise to set default?
        logging.warning("--benchmark-file/-b was not set. Defaulting to use 'benchmark.json'")

    if args.output:
        try:
            makedirs(args.output, exist_ok=True)
            kwargs["output_folder"] = args.output
        except PermissionError:
            sys.exit("Ensure permission exists to create output directory")

    kwargs["rerun"] = args.rerun
    if not args.rerun and isfile(kwargs["benchmark_file"]):
        logging.warning(
            f"Dataset - method combinations whose results are already present in '{kwargs['benchmark_file']}' "
            f"would not be re-run. Use the --rerun flag if this is what is desired.")

    kwargs["is_artifact"] = args.artifact

    classifiers = get_classifiers(args.method, args.determinize)

    if not classifiers:
        sys.exit("Cound not find any valid method - determinization strategy combinations. "
                 "Please read the manual for valid combinations and try again.")

    suite = BenchmarkSuite(**kwargs)
    suite.add_datasets(dataset)
    suite.benchmark(classifiers)


if __name__ == "__main__":
    main()



import dtcontrol.util
from dtcontrol.classifiers.custom_dt import Node
import numpy as np

class OC1Node(Node):
    def __init__(self, coeff, intercept, depth=0):
        super().__init__(depth)
        self.coeff = coeff
        self.intercept = intercept
        self.classifier = None

    def test_condition(self, x):
        """
        :param x: [row of X_train]; shape (1, num_features)
        :return true if go left
        """
        return np.dot(x.flatten(), np.array(self.coeff)) + self.intercept <= 0

    def create_child_node(self):
        pass

    def find_split(self, X, y):
        pass

    def get_dot_label(self):
        if self.actual_label is not None:
            rounded = dtcontrol.util.objround(self.actual_label, 6)  # TODO get precision from flag?
            return dtcontrol.util.split_into_lines(rounded)
        else:
            line = []
            for i in range(len(self.coeff)):
                line.append(f"{round(self.coeff[i], 6)}*X[{i}]")  # TODO get precision from flag?
            line.append(f"{round(self.intercept, 6)}")  # TODO get precision from flag?
            hyperplane = "\n+".join(line) + " <= 0"
            return hyperplane.replace('+-', '-')

    def print_dot_red(self):
        pass

    def is_axis_aligned(self):
        return False



import re
import logging

import numpy as np
import pandas as pd

from dtcontrol.dataset.dataset_loader import DatasetLoader
from tqdm import tqdm


class CSVDatasetLoader(DatasetLoader):
    def _load_dataset(self, filename):
        f = open(filename, 'r')
        print(f"Reading from {filename}")
        f.readline()  # whether permissive
        state_dim, input_dim = map(int, f.readline().split("BEGIN")[1].split())

        ds = pd.read_csv(f, header=None)

        unique_list = []
        for i in range(state_dim, state_dim+input_dim):
            unique_list += ds[i].unique().tolist()
        index_to_value = {x: y for x, y in enumerate(set(unique_list))}
        value_to_index = {y: x for x, y in index_to_value.items()}

        ds[[i for i in range(state_dim, state_dim+input_dim)]] = ds[[i for i in range(state_dim, state_dim+input_dim)]].applymap(lambda x: value_to_index[x])

        grouped = ds.groupby([i for i in range(state_dim)], sort=False)
        aggregate = grouped[state_dim].apply(list).reset_index(name=state_dim)
        for i in range(1, input_dim):
            aggregate[state_dim+i] = grouped[state_dim+i].apply(list).reset_index(name=state_dim+i)[state_dim+i]

        max_non_det = aggregate[state_dim].agg(len).max()

        for i in range(0, input_dim):
            aggregate[state_dim+i] = aggregate[state_dim+i].apply(lambda ls: ls + [-1 for i in range(max_non_det - len(ls))])

        X_train = np.array(aggregate[[i for i in range(state_dim)]])

        if input_dim > 1:
            Y_train = np.ndarray((input_dim, X_train.shape[0], max_non_det), dtype=np.int16)
            for i in range(input_dim):
                Y_train[i] = np.array(aggregate[state_dim+i].tolist())
        else:  # input_dim = 1
            Y_train = np.array(aggregate[state_dim].tolist())

        # construct metadata
        # assumption is that UPPAAL only works with integers
        X_metadata = dict()
        X_metadata["variables"] = [f"x_{i}" for i in range(state_dim)]
        X_metadata["min"] = [int(i) for i in np.amin(X_train, axis=0)]
        X_metadata["max"] = [int(i) for i in np.amax(X_train, axis=0)]
        X_metadata["step_size"] = None  # todo

        Y_metadata = dict()
        Y_metadata["variables"] = [f"u_{i}" for i in range(input_dim)]
        Y_metadata["min"] = [min(index_to_value.values())]
        Y_metadata["max"] = [max(index_to_value.values())]
        Y_metadata["step_size"] = None  # todo

        logging.debug(X_metadata)
        logging.debug(Y_metadata)

        return (X_train, X_metadata, Y_train, Y_metadata, index_to_value)


from dtcontrol.dataset.dataset_loader import DatasetLoader
import numpy as np
import pandas as pd
from os.path import splitext

# This class enables us to load the datasets we have already converted to the XY (vector) format
class VectorDatasetLoader(DatasetLoader):
    # TODO fix this so that it returns X_train, X_metadata, Y_train, Y_metadata, index_to_value
    def _load_dataset(self, filename):
        path, _ = splitext(filename)
        X_train = pd.read_pickle(filename)
        Y_train = np.load(f'{path}.npy')
        Y_train = self.convert_vector_labels_to_new_format(Y_train)
        return np.array(X_train), list(X_train.columns), Y_train, {x:x for x in Y_train.flatten()}

    @staticmethod
    def convert_vector_labels_to_new_format(labels):
        m = np.max(np.nonzero(labels)[1]) + 1
        l = []
        for row in labels:
            values = list(np.nonzero(row)[0] + 1)
            filler = [-1] * (m - len(values))
            l.append(values + filler)
        return np.array(l)



from dtcontrol.classifiers.linear_classifier_dt import LinearClassifierDT, LinearClassifierOrAxisAlignedNode
from dtcontrol.dataset.multi_output_dataset import MultiOutputDataset

class MaxFreqMultiLinearClassifierDT(LinearClassifierDT):
    def __init__(self, classifier_class, **kwargs):
        super().__init__(classifier_class, **kwargs)
        self.name = 'MaxFreqMulti-LinearClassifierDT'

    def is_applicable(self, dataset):
        return isinstance(dataset, MultiOutputDataset) and not dataset.is_deterministic

    def fit(self, dataset):
        if dataset.tuple_to_tuple_id is None:
            dataset.get_tuple_ids()
        self.root = MaxFreqMultiLinearClassifierNode(dataset.tuple_to_tuple_id, self.classifier_class, **self.kwargs)
        self.root.fit(dataset.X_train, dataset.Y_train)
        self.set_labels(lambda leaf: dataset.map_tuple_id_back(leaf.trained_label), dataset.index_to_value)

class MaxFreqMultiLinearClassifierNode(LinearClassifierOrAxisAlignedNode):
    def __init__(self, tuple_to_tuple_id, classifier_class, depth=0, **kwargs):
        super().__init__(classifier_class, depth, **kwargs)
        self.tuple_to_tuple_id = tuple_to_tuple_id

    def create_child_node(self):
        return MaxFreqMultiLinearClassifierNode(self.tuple_to_tuple_id, self.classifier_class, self.depth + 1,
                                                **self.kwargs)

    def find_split(self, X, y):
        return super().find_split(X, MultiOutputDataset.determinize_max_over_all_inputs(y, self.tuple_to_tuple_id))

    def check_done(self, X, y):
        return super().check_done(X, MultiOutputDataset.determinize_max_over_all_inputs(y, self.tuple_to_tuple_id))



import pickle
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable
import dtcontrol
import numpy as np
from sklearn.base import BaseEstimator

from jinja2 import FileSystemLoader, Environment

file_loader = FileSystemLoader([path + "/c_templates" for path in dtcontrol.__path__])
env = Environment(loader=file_loader)
single_output_c_template = env.get_template('single_output.c')
multi_output_c_template = env.get_template('multi_output.c')


class CustomDT(ABC, BaseEstimator):
    def __init__(self):
        self.root = None
        self.name = None

    @abstractmethod
    def is_applicable(self, dataset):
        pass

    def set_labels(self, leaf_fun, index_to_value):
        self.root.set_labels(leaf_fun, index_to_value)

    @abstractmethod
    def fit(self, dataset):
        pass

    def predict(self, dataset):
        return self.root.predict(dataset.X_train)

    def get_stats(self) -> dict:
        return {
            'nodes': self.root.num_nodes,
            'bandwidth': int(np.ceil(np.log2((self.root.num_nodes + 1) / 2)))
        }

    def export_dot(self, file=None):
        dot = self.root.export_dot()
        if file:
            with open(file, 'w+') as outfile:
                outfile.write(dot)
        else:
            return dot

    def export_c(self, num_outputs, example, file=None):
        template = multi_output_c_template if num_outputs > 1 else single_output_c_template
        code = self.root.export_c()
        result = template.render(example=example, num_outputs=num_outputs, code=code)
        if file:
            with open(file, 'w+') as outfile:
                outfile.write(result)
        else:
            return result

    # Needs to know the number of inputs, because it has to define how many inputs the hardware component has in the "entity" block
    def export_vhdl(self, numInputs, file=None):
        entitystr = "entity controller is\n\tport (\n"
        allInputs = ""
        for i in range(0, numInputs):
            entitystr += "\t\tx" + str(i) + ": in <type>;\n"  # todo: get type from dtcontrol.dataset :(
            allInputs += "x" + str(i) + ","
        entitystr += "\t\ty: out <type>\n\t);\nend entity;\n\n"  # no semicolon after last declaration. todo: multi-output; probably just give dataset to this method...
        architecture = "architecture v1 of controller is\nbegin\n\tprocess(" + allInputs[
                                                                               :-1] + ")\n\tbegin\n" + self.root.export_vhdl() + "\n\tend process;\nend architecture;"
        if file:
            with open(file, 'w+') as outfile:
                outfile.write(entitystr + architecture)
        else:
            return entitystr + architecture

    def save(self, filename):
        with open(filename, 'wb') as outfile:
            pickle.dump(self, outfile)

    def __str__(self):
        return self.name

class Node(ABC):
    def __init__(self, depth=0):
        self.left = None
        self.right = None
        self.trained_label = None  # the label from the training data the node sees (e.g. unique)
        self.mapped_label = None  # the label corresponding to the actual int labels
        self.actual_label = None  # the actual float label
        self.depth = depth
        self.num_nodes = 0

    def set_labels(self, leaf_fun, index_to_value):
        def _visit_leaves(tree):
            if tree is None:
                return
            if tree.trained_label is not None:
                tree.mapped_label = leaf_fun(tree)
                # the mapped label can be either a list of labels or a single label
                if isinstance(tree.mapped_label, tuple):
                    tree.actual_label = tuple([index_to_value[i] for i in tree.mapped_label if i != -1])
                elif isinstance(tree.mapped_label, Iterable):
                    if isinstance(tree.mapped_label[0], tuple):
                        tree.actual_label = [tuple(map(lambda x: index_to_value[x], tup)) for tup in tree.mapped_label]
                    else:
                        tree.actual_label = [index_to_value[i] for i in tree.mapped_label if i != -1]
                else:
                    tree.actual_label = index_to_value[tree.mapped_label]
            _visit_leaves(tree.left)
            _visit_leaves(tree.right)

        _visit_leaves(self)

    def predict(self, X):
        pred = []
        for row in np.array(X):
            pred.append(self.predict_one(row.reshape(1, -1)))
        return pred

    def predict_one(self, features):
        node = self
        while node.left:
            node = node.left if node.test_condition(features) else node.right
        return node.mapped_label

    @abstractmethod
    def test_condition(self, x):
        pass

    def fit(self, X, y):
        if self.check_done(X, y):
            return
        mask = self.find_split(X, y)
        self.left = self.create_child_node()
        self.right = self.create_child_node()
        self.fit_children(X, y, mask)

    @abstractmethod
    def create_child_node(self):
        pass

    def check_done(self, X, y):
        if self.depth >= 500:
            print("Cannot find a good split.")
            return True

        unique_labels = np.unique(y)
        num_unique_labels = len(unique_labels)
        unique_data = np.unique(X, axis=0)
        num_unique_data = len(unique_data)
        if num_unique_labels <= 1 or num_unique_data <= 1:
            self.trained_label = y[0] if len(unique_labels) > 0 else None
            self.num_nodes = 1
            return True
        return False

    @abstractmethod
    def find_split(self, X, y):
        pass

    def fit_children(self, X, y, mask):
        if len(y.shape) == 3:
            left_labels, right_labels = y[:, mask, :], y[:, ~mask, :]
        else:
            left_labels, right_labels = y[mask], y[~mask]
        self.left.fit(X[mask], left_labels)
        self.right.fit(X[~mask], right_labels)
        self.num_nodes = 1 + self.left.num_nodes + self.right.num_nodes

    @staticmethod
    def calculate_impurity(labels, mask):
        left = labels[mask]
        right = labels[~mask]
        if len(left) == 0 or len(right) == 0: return sys.maxsize
        num_labels = len(labels)
        return (len(left) / num_labels) * Node.calculate_entropy(left) + \
               (len(right) / num_labels) * Node.calculate_entropy(right)

    @staticmethod
    def calculate_entropy(labels):
        num_labels = len(labels)
        unique = np.unique(labels)
        probs = [len(labels[labels == label]) / num_labels for label in unique]
        return sum(-prob * np.log2(prob) for prob in probs)

    def export_dot(self):
        text = 'digraph {{\n{}\n}}'.format(self._export_dot(0)[1])
        return text

    def _export_dot(self, starting_number):
        if not self.left and not self.right:
            return starting_number, '{} [label=\"{}\"];\n'.format(starting_number, self.get_dot_label())

        text = '{} [label=\"{}\"'.format(starting_number, self.get_dot_label())
        if self.print_dot_red():
            text += ', fillcolor=firebrick1, style=filled'
        text += "];\n"

        number_for_right = starting_number + 1
        last_number = starting_number

        if self.left:
            last_left_number, left_text = self.left._export_dot(starting_number + 1)
            text += left_text
            label = 'True' if starting_number == 0 else ''
            text += '{} -> {} [label="{}"];\n'.format(starting_number, starting_number + 1, label)
            number_for_right = last_left_number + 1
            last_number = last_left_number

        if self.right:
            last_right_number, right_text = self.right._export_dot(number_for_right)
            text += right_text
            label = 'False' if starting_number == 0 else ''
            text += '{} -> {} [style="dashed", label="{}"];\n'.format(starting_number, number_for_right, label)
            last_number = last_right_number

        return last_number, text

    @abstractmethod
    def is_axis_aligned(self):
        pass

    def export_c(self):
        return self._export_if_then_else(0, 'c')

    def export_vhdl(self):
        return self._export_if_then_else(2, 'vhdl')

    # This handles both the c and vhdl export of nodes, as they both have the if-then-else structure and differ only in details
    # Type can be 'c' or 'vhdl', currently
    def _export_if_then_else(self, indent_index, type):
        # get label depending on type
        if type == 'c':
            label = self.get_c_label()
        elif type == 'vhdl':
            label = self.get_vhdl_label()
        else:
            raise Exception(
                f'Unknown type {type} in _export_if_then_else')  # do this type check once; from now on, else always means vhdl
        # If leaf node
        if not self.left and not self.right:
            return "\t" * indent_index + str(label)

        # if inner node
        text = ""
        text += "\t" * indent_index + (
            f"if ({self.get_c_label()}) {{\n" if type == 'c' else f"if {self.get_vhdl_label()} then\n")

        if self.left:
            text += f"{self.left._export_if_then_else(indent_index + 1, type)}\n"
        else:
            text += "\t" * (indent_index + 1) + ";\n"
        if type == 'c':
            text += "\t" * indent_index + "}\n"

        if self.right:
            text += "\t" * indent_index + ("else {\n" if type == 'c' else "else \n")
            text += f"{self.right._export_if_then_else(indent_index + 1, type)}\n"
            text += "\t" * indent_index + ("}" if type == 'c' else "end if;")

        return text

    def get_determinized_label(self):
        if isinstance(self.actual_label,
                      list):  # list of things, i.e. nondeterminsm present, need to determinise by finding max norm (could use other ideas as well)
            return min(self.actual_label, key=lambda l: self.norm(l))
        else:  # just an int or a tuple, easy; note that we have to use list and not iterable in the above if, because tuples (multi control input) are also iterable
            return self.actual_label

            # If tup is an int, return abs of that number; if it is a tuple, return sum of absolutes of contents

    # Used by get_determinized_label
    def norm(self, tup):
        if isinstance(tup, tuple):
            return sum([i * i for i in tup])
        else:  # if it is not iterable, then it was a single number, return abs of that
            return abs(tup)

    @abstractmethod
    def get_dot_label(self):
        pass

    @abstractmethod
    def print_dot_red(self):
        pass

    def get_c_label(self):
        return self._get_label('c')

    def get_vhdl_label(self):
        return self._get_label('vhdl')

    def _get_label(self, type):
        # if leaf, return class; determinize if necessary, and handle multi output printing
        if self.actual_label is not None:
            label = self.get_determinized_label()
            if isinstance(label, Iterable):
                if type == 'c':
                    return ' '.join([
                        f'result[{i}] = {round(label[i], 6)}f;' for i in range(len(label))
                    ])
                else:
                    i = 0
                    result = ""
                    for controlInput in label:
                        result += f'y{str(i)} <= {str(controlInput)}; '
                        i += 1
                    return result
            else:
                if type == 'c':
                    return f'return {round(float(label), 6)}f;'
                else:
                    return f'y <= {str(label)};'
        # else, print predicate in correct format

        if self.is_axis_aligned():
            try:  # the guy keeping track of the tree is called differently depending on subclass. This ain't nice, but it works.
                tree = self.dt.tree_
            except:
                tree = self.classifier.tree_
            l = f'X[{tree.feature[0]}]' if type == 'c' else f'x{tree.feature[0]}'
            return f'{l} <= {round(tree.threshold[0], 4)}'
        else:
            # this implicitly assumes n_classes == 2
            if self.classifier is not None:
                coef_ = self.classifier.coef_[0]
                intercept_ = self.classifier.intercept_[0]
            else:  # OC1Node
                coef_ = self.coeff
                intercept_ = self.intercept
            line = []
            for i in range(0, len(coef_)):
                line.append(f"{coef_[i]}*X[{i}]" if type == 'c' else f"{coef_[i]}*x{i}")
            line.append(f"{intercept_}")
            hyperplane = "+".join(line) + " <= 0"
            return hyperplane.replace('+-', '-')



import dtcontrol.util
from sklearn.tree import DecisionTreeClassifier
from dtcontrol.classifiers.custom_dt import CustomDT, Node

class CartDT(CustomDT):
    def __init__(self):
        super().__init__()
        self.name = 'CART'

    def is_applicable(self, dataset):
        return True

    def fit(self, dataset):
        self.root = CartNode()
        self.root.fit(dataset.X_train, dataset.get_unique_labels())
        self.set_labels(lambda leaf: dataset.map_unique_label_back(leaf.trained_label), dataset.index_to_value)

class CartNode(Node):
    def __init__(self, depth=0):
        super().__init__(depth)
        self.dt = None

    def test_condition(self, x):
        tree = self.dt.tree_
        return x[:, tree.feature[0]][0] <= tree.threshold[0]

    def create_child_node(self):
        return CartNode(self.depth + 1)

    def find_split(self, X, y):
        self.dt = DecisionTreeClassifier(max_depth=1, criterion='entropy')
        self.dt.fit(X, y)
        mask = X[:, self.dt.tree_.feature[0]] <= self.dt.tree_.threshold[0]
        return mask

    def get_dot_label(self):
        if self.actual_label is not None:
            rounded = dtcontrol.util.objround(self.actual_label, 6)  # TODO get precision from flag?
            return dtcontrol.util.split_into_lines(rounded)
        tree = self.dt.tree_
        return f'X[{tree.feature[0]}] <= {round(tree.threshold[0], 6)}'  # TODO get precision from flag?

    def print_dot_red(self):
        return False
        
    def is_axis_aligned(self):
        return True



from dtcontrol.classifiers.cart_custom_dt import CartDT, CartNode
from dtcontrol.dataset.dataset import Dataset
from dtcontrol.dataset.single_output_dataset import SingleOutputDataset

class MaxFreqDT(CartDT):
    def __init__(self, ):
        super().__init__()
        self.name = 'MaxFreqDT'

    def is_applicable(self, dataset):
        return not dataset.is_deterministic

    def fit(self, dataset):
        self.root = MaxFreqNode()
        if isinstance(dataset, SingleOutputDataset):
            self.root.fit(dataset.X_train, dataset.Y_train)
            self.set_labels(lambda leaf: leaf.trained_label, dataset.index_to_value)
        else:
            self.root.fit(dataset.X_train, dataset.get_tuple_ids())
            self.set_labels(lambda leaf: dataset.map_tuple_id_back(leaf.trained_label), dataset.index_to_value)

class MaxFreqNode(CartNode):
    def __init__(self, depth=0):
        super().__init__(depth)

    def create_child_node(self):
        return MaxFreqNode(self.depth + 1)

    def find_split(self, X, y):
        return super().find_split(X, Dataset._get_max_labels(y))

    def check_done(self, X, y):
        return super().check_done(X, Dataset._get_max_labels(y))



import numpy as np
import pandas as pd

filename = '../XYdatasets/aircraft'
first_y_index = 3

data = np.genfromtxt(f'{filename}.scots', delimiter=',')
data = data[1:]  # remove column names
assert len(data) == len(np.unique(data[:, :first_y_index], axis=1))  # should be non-deterministic

X = data[:, :first_y_index]
Y = data[:, first_y_index:]
data_frame = pd.DataFrame(np.array(X), columns=[f'x{i+1}' for i in range(first_y_index)])
data_frame.to_pickle(f'{filename}_X.pickle')
np.save(f'{filename}_Y.npy', np.array(Y))



import sys
import os

"""
This adds the parent folder to the python path in order to be able to import modules from other folders 
(e.g. smarter_splits).
"""

module_path = os.path.abspath(os.path.join(os.pardir))
if module_path not in sys.path:
    sys.path.append(module_path)


from dtcontrol.dataset.dataset import Dataset
from dtcontrol.dataset.single_output_dataset import SingleOutputDataset
from dtcontrol.classifiers.linear_classifier_dt import LinearClassifierDT, LinearClassifierOrAxisAlignedNode

class MaxFreqLinearClassifierDT(LinearClassifierDT):
    def __init__(self, classifier_class, **kwargs):
        super().__init__(classifier_class, **kwargs)
        self.name = 'MaxFreq-LinearClassifierDT-{}'.format(classifier_class.__name__)

    def is_applicable(self, dataset):
        return not dataset.is_deterministic

    def fit(self, dataset):
        self.root = MaxFreqLinearClassifierNode(self.classifier_class, **self.kwargs)
        if isinstance(dataset, SingleOutputDataset):
            self.root.fit(dataset.X_train, dataset.Y_train)
            self.set_labels(lambda leaf: leaf.trained_label, dataset.index_to_value)
        else:
            self.root.fit(dataset.X_train, dataset.get_tuple_ids())
            self.set_labels(lambda leaf: dataset.map_tuple_id_back(leaf.trained_label), dataset.index_to_value)

class MaxFreqLinearClassifierNode(LinearClassifierOrAxisAlignedNode):
    def __init__(self, classifier_class, depth=0, **kwargs):
        super().__init__(classifier_class, depth, **kwargs)

    def create_child_node(self):
        return MaxFreqLinearClassifierNode(self.classifier_class, self.depth + 1, **self.kwargs)

    def find_split(self, X, y):
        return super().find_split(X, Dataset._get_max_labels(y))

    def check_done(self, X, y):
        return super().check_done(X, Dataset._get_max_labels(y))



import pickle
from abc import ABC, abstractmethod
from os import makedirs
from os.path import getmtime, split, exists, join

import numpy as np
import pandas as pd

class DatasetLoader(ABC):
    def __init__(self):
        self.loaded_datasets = {}
        self.PATH = '.benchmark_suite'

    def load_dataset(self, filename):
        if filename not in self.loaded_datasets:
            if self.already_converted(filename):
                self.loaded_datasets[filename] = self.load_converted_dataset(filename)
            else:
                self.loaded_datasets[filename] = self._load_dataset(filename)
                self.save_converted_dataset(filename)
        return self.loaded_datasets[filename]

    def already_converted(self, filename):
        possible_path = self.get_converted_folder(filename)
        return exists(possible_path)

    def load_converted_dataset(self, filename):
        folder = self.get_converted_folder(filename)
        assert exists(folder)
        print("Loading existing pickled dataset...", end=' ')
        X_train = np.load(join(folder, 'X_train.npy'))
        Y_train = np.load(join(folder, 'Y_train.npy'))
        with open(join(folder, 'extra_data.pickle'), 'rb') as infile:
            extra_data = pickle.load(infile)
            index_to_value = extra_data["index_to_value"]
            X_metadata = extra_data["X_metadata"]
            Y_metadata = extra_data["Y_metadata"]
        print("done loading.")
        return X_train, X_metadata, Y_train, Y_metadata, index_to_value

    def save_converted_dataset(self, filename):
        folder = self.get_converted_folder(filename)
        if not exists(folder):
            makedirs(folder)
        X_train, X_metadata, Y_train, Y_metadata, index_to_value = self.loaded_datasets[filename]
        np.save(join(folder, 'X_train.npy'), X_train)
        np.save(join(folder, 'Y_train.npy'), Y_train)
        with open(join(folder, 'extra_data.pickle'), 'wb+') as outfile:
            pickle.dump({"index_to_value": index_to_value,
                         "X_metadata": X_metadata,
                         "Y_metadata": Y_metadata},
                        outfile)

    def get_converted_folder(self, filename):
        directory, name = split(filename)
        return join(directory, self.PATH, f'{name}_{getmtime(filename)}')

    """
    Loads a dataset and returns X_train, X_metadata, Y_train, Y_metadata, index_to_value
    """

    @abstractmethod
    def _load_dataset(self, filename):
        pass



import numpy as np
import pandas as pd

filename = '../XYdatasets/cartpole'

data = np.genfromtxt(f'{filename}.scots', delimiter=',')
data = data[1:]  # remove column names
# nondeterministic = len(data) - len(np.unique(data[:,:first_y_index], axis=1)) > 0  # only true for cartpole up to now

labels = np.unique(data[:,2])
label_mapping = dict()
for i in range(len(labels)):
    label_mapping[labels[i]] = i

same_points_indices = [0]
i = 0
while i < len(data):
    j = i
    while j < len(data) - 1 and (data[j,:2] == data[j+1,:2]).all():
        j += 1
    same_points_indices.append(j + 1)
    i = j + 1

new_data = []
new_labels = []
for i in range(len(same_points_indices) - 1):
    new_data.append(data[same_points_indices[i],:2])
    label_indices = [label_mapping[l] for l in data[same_points_indices[i]:same_points_indices[i+1],2]]
    label = np.array([0] * len(labels))
    label[label_indices] = 1
    new_labels.append(label)

data_frame = pd.DataFrame(np.array(new_data), columns=['x1','x2'])
data_frame.to_pickle(f'{filename}_X.pickle')

np.save(f'{filename}_Y.npy', np.array(new_labels))


import glob
import json
import logging
import os
import time
import webbrowser
import dtcontrol
from os import makedirs
from os.path import join, exists, isfile

from dtcontrol.classifiers.oc1_wrapper import OC1Wrapper
from dtcontrol.dataset.multi_output_dataset import MultiOutputDataset
from dtcontrol.dataset.single_output_dataset import SingleOutputDataset
from dtcontrol.timeout import call_with_timeout
from dtcontrol.ui.table_controller import TableController
from dtcontrol.util import format_seconds, get_filename_and_ext


class BenchmarkSuite:
    """
    The benchmark suite runs the given classifiers on all datasets present in the XYdatasets folder and prints the
    results.

    The classifiers have to satisfy the following interface:
        classifier.name returns the name to be displayed in the results
        classifier.fit(dataset) trains the classifier on the given dataset
        classifier.predict(dataset) returns an array of the classifiers predictions for the dataset
        classifier.get_stats() returns a dictionary of statistics to be displayed (e.g. the number of nodes in the tree)
        classifier.is_applicable(dataset) returns whether the classifier can be applied to the dataset
        classifier.save(file) saves the classifier to a file (for debugging purposes only)
        classifier.export_dot(file) saves a dot-representation of the classifier to a file
        classifier.export_c(file) saves a C-representation of the classifier to a file
    """

    def __init__(self, benchmark_file='benchmark', timeout=None, output_folder='decision_trees', save_folder=None,
                 rerun=False, use_multiprocessing=True, is_artifact=False):
        logging.basicConfig(level=logging.INFO, format='%(message)s')
        self.datasets = []
        self.json_file = f'{benchmark_file}.json'
        self.html_file = f'{benchmark_file}.html'
        self.results = {}
        self.timeout = timeout
        self.output_folder = output_folder
        self.save_folder = save_folder
        self.rerun = rerun
        self.use_multiprocessing = use_multiprocessing
        self.is_artifact = is_artifact
        self.table_controller = TableController(self.html_file, self.output_folder, self.is_artifact)

        logging.info(f"Benchmark statistics will be available in {self.json_file} and {self.html_file}.")
        logging.info(f"Constructed trees will be written to {self.output_folder}.")

    def add_datasets(self, paths, include=None, exclude=None):
        if not exclude:
            exclude = []
        if include is not None and len(set(include) & set(exclude)) > 0:
            logging.error('A dataset cannot be both included and excluded.\nAborting.')
            return
        self.datasets = []
        for path in paths:
            for file in self.get_files(path):
                name, ext = get_filename_and_ext(file)
                if ((not include) or name in include) and name not in exclude:
                    if self.is_multiout(file, ext):
                        ds = MultiOutputDataset(file)
                    else:
                        ds = SingleOutputDataset(file)

                    # check if dataset is deterministic
                    ds.is_deterministic = self.is_deterministic(file, ext)
                    self.datasets.append(ds)
        self.datasets.sort(key=lambda ds: ds.name)

    def get_files(self, path):
        if isfile(path):
            return [path]
        else:
            return glob.glob(join(path, '*.scs')) + glob.glob(join(path, '*.dump'))

    def display_html(self):
        # display(HTML(f'<html><a href="{self.html_file}" target="_blank">View table</a></html>'))
        url = f'file://{os.path.abspath(self.html_file)}'
        webbrowser.open(url)

    def benchmark(self, classifiers):
        self.load_results()
        num_steps = len(classifiers) * len(self.datasets)
        step = 0
        for ds in self.datasets:
            for classifier in classifiers:
                step += 1
                logging.info(f"{step}/{num_steps}: Evaluating {classifier.name} on {ds.name}... ")
                cell, computed = self.compute_cell(ds, classifier)
                if computed:
                    self.save_result(classifier.name, ds, cell)
                    if cell == 'timeout':
                        msg = f"{step}/{num_steps}: {classifier.name} on {ds.name} timed out after {format_seconds(self.timeout)}"
                    else:
                        msg = f"{step}/{num_steps}: Evaluated {classifier.name} on {ds.name} in {cell['time']}."
                    logging.info(msg)
                else:
                    if cell == 'not applicable':
                        self.save_result(classifier.name, ds, cell)
                        logging.info(f"{step}/{num_steps}: {classifier.name} is not applicable for {ds.name}.")
                    else:
                        logging.info(
                            f"{step}/{num_steps}: Not running {classifier.name} on {ds.name} as result available in {self.json_file}.")
        logging.info('All benchmarks completed. Shutting down dtControl.')

        self.table_controller.update_and_save(self.results, [ds.name for ds in self.datasets],
                                              [cl.name for cl in classifiers])

    def compute_cell(self, dataset, classifier):
        if self.already_computed(dataset, classifier) and not self.rerun:
            computed = False
            cell = self.results[dataset.name]['classifiers'][classifier.name]
        elif not classifier.is_applicable(dataset):
            computed = False
            cell = 'not applicable'
        else:
            computed = True
            dataset.load_if_necessary()
            cell = self.train_and_get_cell(dataset, classifier)
        return cell, computed

    def already_computed(self, dataset, classifier):
        return dataset.name in self.results and classifier.name in self.results[dataset.name]['classifiers']

    def train_and_get_cell(self, dataset, classifier):
        if self.timeout is not None:
            classifier, success, run_time = call_with_timeout(classifier, 'fit', dataset, timeout=self.timeout)
        else:
            start = time.time()
            classifier.fit(dataset)
            run_time = time.time() - start
            success = True
        if success:
            # todo once BDD is working for all models
            # if isinstance(classifier, BDD):  # bdd cannot predict yet
            #     stats = classifier.get_stats()
            #     cell = {'stats': stats, 'time': format_seconds(run_time)}
            #     return cell
            acc = dataset.compute_accuracy(classifier.predict(dataset))
            if acc is None:
                cell = 'failed to fit'
            else:
                stats = classifier.get_stats()
                cell = {'stats': stats, 'time': format_seconds(run_time)}
                dot_filename = self.get_filename(self.output_folder, dataset, classifier, '.dot')
                classifier.export_dot(dot_filename)
                c_filename = self.get_filename(self.output_folder, dataset, classifier, '.c')
                num_outputs = 1 if len(dataset.Y_train.shape) <= 2 else len(dataset.Y_train)
                classifier.export_c(num_outputs, f'{{{",".join(str(i) + "f" for i in dataset.X_train[0])}}}', c_filename)
                if not isinstance(classifier, OC1Wrapper):  # TODO vhdl for OC1? not until artifact deadline
                    vhdl_filename = self.get_filename(self.output_folder, dataset, classifier, '.vhdl')
                    classifier.export_vhdl(len(dataset.X_metadata["variables"]), vhdl_filename)
                if abs(acc - 1.0) > 1e-10:
                    cell['accuracy'] = acc
                if self.save_folder is not None:
                    classifier.save(self.get_filename(self.save_folder, dataset, classifier, '.saved', unique=True))
        else:
            cell = 'timeout'
        return cell

    @staticmethod
    def get_filename(folder, dataset, classifier, extension, unique=False):
        dir = join(folder, classifier.name, dataset.name)
        if not exists(dir):
            makedirs(dir)
        name = classifier.name
        if unique:
            name += f'--{time.strftime("%Y%m%d-%H%M%S")}'
        name += extension
        return join(dir, name)

    def save_result(self, classifier_name, dataset, result):
        if dataset.name not in self.results:
            self.results[dataset.name] = {'classifiers': {},
                                          'metadata': {
                                              'X_metadata': dataset.X_metadata,
                                              'Y_metadata': dataset.Y_metadata
                                          }
                                          }
        self.results[dataset.name]['classifiers'][classifier_name] = result
        self.save_to_disk()

    def load_results(self):
        if not self.json_file or not exists(self.json_file) or not isfile(self.json_file):
            return
        with open(self.json_file, 'r') as infile:
            self.results = json.load(infile)
        for ds in self.datasets:
            if ds.name in self.results:
                ds.load_metadata_from_json(self.results[ds.name])

    def delete_dataset_results(self, dataset_name):
        self.load_results()
        if dataset_name in self.results:
            del self.results[dataset_name]
        self.save_to_disk()

    def delete_classifier_results(self, classifier_name):
        self.load_results()
        for dataset in self.results:
            classifiers = self.results[dataset]['classifiers']
            if classifier_name in classifiers:
                del classifiers[classifier_name]
        self.save_to_disk()

    def delete_cell(self, dataset_name, classifier_name):
        self.load_results()
        if dataset_name in self.results:
            classifiers = self.results[dataset_name]['classifiers']
            if classifier_name in classifiers:
                del classifiers[classifier_name]
        self.save_to_disk()

    def save_to_disk(self):
        with open(self.json_file, 'w+') as outfile:
            json.dump(self.results, outfile, indent=2)

    @staticmethod
    def is_multiout(filename, ext):
        if "scs" in ext:
            # if scs, then
            f = open(filename)
            # Read input dim from scs file
            for i in range(5):
                f.readline()
            state_dim = int(f.readline())
            for i in range(12 + 3 * state_dim):
                f.readline()
            input_dim = int(f.readline())
            return input_dim > 1
        elif "csv" in ext:
            # if scs, then
            f = open(filename)
            f.readline()
            _, input_dim = map(int, f.readline().split("BEGIN")[1].split())
            return input_dim > 1
        else:
            return False

    @staticmethod
    def is_deterministic(filename, ext):
        if "scs" in ext:
            f = open(filename)
            # Read input dim from scs file
            for i in range(5):
                f.readline()
            state_dim = int(f.readline())
            for i in range(12 + 3 * state_dim):
                f.readline()
            input_dim = int(f.readline())
            for i in range(12 + 3 * input_dim):
                f.readline()
            non_det = int(f.readline().split(":")[1].split()[1])
            return non_det == 1
        elif "csv" in ext:
            f = open(filename)
            is_det = "NON-PERMISSIVE" in f.readline()
            return is_det
        else:
            return False  # UPPAAL is always non-deterministic



import numpy as np

from dtcontrol.classifiers.cart_custom_dt import CartDT, CartNode
from dtcontrol.dataset.single_output_dataset import SingleOutputDataset

class NormDT(CartDT):
    """
    :param comp: The comparison function to be used (either max or min)
    """

    def __init__(self, comp):
        super().__init__()
        self.name = f'{"Max" if comp == max else "Min"}NormDT'
        self.comp = comp

    def is_applicable(self, dataset):
        return not dataset.is_deterministic

    def fit(self, dataset):
        self.root = CartNode()
        if isinstance(dataset, SingleOutputDataset):
            self.root.fit(dataset.X_train, self.determinize_single_output(dataset))
            self.set_labels(lambda leaf: leaf.trained_label, dataset.index_to_value)
        else:
            self.root.fit(dataset.X_train, self.determinize_multi_output(dataset))
            self.set_labels(lambda leaf: dataset.map_tuple_id_back(leaf.trained_label), dataset.index_to_value)

    def determinize_single_output(self, dataset):
        return np.apply_along_axis(lambda x: self.comp(x[x != -1], key=lambda i: dataset.index_to_value[i] ** 2),
                                   axis=1,
                                   arr=dataset.Y_train)

    def determinize_multi_output(self, dataset):
        zipped = np.stack(dataset.Y_train, axis=2)
        result = []
        i = 0
        for arr in zipped:
            result.append(self.comp([t for t in arr if t[0] != -1],
                                    key=lambda t: sum(dataset.index_to_value[i] ** 2 for i in t)))
            i += 1
        if dataset.tuple_to_tuple_id is None:  # TODO: refactor stuff like this with something like dataset.get_tuple_to_tuple_ids()
            dataset.get_unique_labels()
        return np.array([dataset.tuple_to_tuple_id[tuple(t)] for t in result])



import logging

import numpy as np
from dtcontrol.dataset.dataset_loader import DatasetLoader
from tqdm import tqdm


class ScotsDatasetLoader(DatasetLoader):
    def _load_dataset(self, filename):
        precision = 10

        f = open(filename)
        print(f"Reading from {filename}")

        for i in range(5):
            f.readline()

        line = f.readline()
        state_dim = int(line)

        for i in range(2):
            f.readline()

        state_eta = []
        for i in range(state_dim):
            line = f.readline()
            state_eta.append(round(float(line), precision))

        for i in range(3):
            f.readline()

        state_lb = []
        for i in range(state_dim):
            line = f.readline()
            state_lb.append(round(float(line), precision))

        for i in range(3):
            f.readline()

        state_ub = []
        for i in range(state_dim):
            line = f.readline()
            state_ub.append(round(float(line), precision))

        n_state_grid = []
        for i in range(state_dim):
            n_state_grid.append(int(round((state_ub[i] - state_lb[i]) / state_eta[i], precision)) + 1)

        for i in range(4):
            f.readline()

        input_dim = int(f.readline())

        for i in range(2):
            f.readline()

        input_eta = []
        for i in range(input_dim):
            line = f.readline()
            input_eta.append(round(float(line), precision))

        for i in range(3):
            f.readline()

        input_lb = []
        for i in range(input_dim):
            line = f.readline()
            input_lb.append(round(float(line), precision))

        for i in range(3):
            f.readline()

        input_ub = []
        for i in range(input_dim):
            line = f.readline()
            input_ub.append(float(line))

        n_input_grid = []
        for i in range(input_dim):
            n_input_grid.append(int(round((input_ub[i] - input_lb[i]) / input_eta[i], precision) + 1))

        for i in range(4):
            f.readline()

        dim_str = f.readline().split(":")[1]
        rows, max_non_det = list(map(int, dim_str.split()))

        # Controller starts now
        controller_start = 32 + 3 * (state_dim + input_dim)

        x_NN = [1]
        for i in range(1, state_dim):
            x_NN.append(x_NN[i - 1] * n_state_grid[i - 1])

        u_NN = [1]
        for i in range(1, input_dim):
            u_NN.append(u_NN[i - 1] * n_input_grid[i - 1])

        # Get the number of lines describing the controller
        controller_lines = sum(1 for line in f) - 1

        x_train = np.empty((controller_lines, state_dim), dtype=np.float32)
        y_train = np.full((input_dim, controller_lines, max_non_det), -1, dtype=np.int32)

        f.seek(0)

        float_to_unique_label = dict()

        for i in range(controller_start):
            f.readline()

        logging.info("Extracting states and control inputs from SCOTS dump")
        for i, line in enumerate(tqdm(f, total=controller_lines)):
            if i == controller_lines:
                break
            idxu = np.fromstring(line, dtype=np.int32, sep=' ')
            idx = idxu[0]

            k = state_dim - 1
            x = np.zeros(state_dim)
            while k > 0:
                num = int(idx / x_NN[k])  # j
                idx = idx % x_NN[k]  # i
                x[k] = state_lb[k] + num * state_eta[k]
                k = k - 1
            num = idx
            x[0] = state_lb[0] + num * state_eta[0]

            x_train[i] = x

            # creating input variables
            u_idx = np.empty(input_dim, dtype=np.int16)
            for j in range(1, idxu.shape[0]):
                idu = idxu[j]
                kk = input_dim - 1
                u = np.empty(input_dim, dtype=np.float32)
                while kk > 0:
                    u_idx[kk] = int(idu / u_NN[kk])
                    idu = idu % u_NN[kk]
                    u[kk] = input_lb[kk] + u_idx[kk] * input_eta[kk]
                    if u[kk] not in float_to_unique_label.keys():
                        float_to_unique_label[u[kk]] = len(float_to_unique_label) + 1
                    y_train[kk][i][j - 1] = float_to_unique_label[u[kk]]
                    kk = kk - 1
                u_idx[0] = idu
                u[0] = input_lb[0] + u_idx[0] * input_eta[0]
                if u[0] not in float_to_unique_label.keys():
                    float_to_unique_label[u[0]] = len(float_to_unique_label) + 1
                y_train[0][i][j - 1] = float_to_unique_label[u[0]]

        # inverse map
        unique_label_to_float = {y: x for (x, y) in float_to_unique_label.items()}

        # if only single control input, do not wrap it in another array
        if y_train.shape[0] == 1:
            y_train = y_train[0]

        print("Constructed training set with %s datapoints" % x_train.shape[0])

        # construct metadata
        x_metadata = dict()
        x_metadata["variables"] = [f"x_{i}" for i in range(len(state_lb))]
        x_metadata["min"] = state_lb
        x_metadata["max"] = state_ub
        x_metadata["step_size"] = state_eta

        y_metadata = dict()
        y_metadata["variables"] = [f"u_{i}" for i in range(len(input_lb))]
        y_metadata["min"] = input_lb
        y_metadata["max"] = input_ub
        y_metadata["step_size"] = input_eta

        return (x_train, x_metadata, y_train, y_metadata, unique_label_to_float)



import numpy as np
from operator import itemgetter
from dtcontrol.dataset.dataset import Dataset
from dtcontrol.util import make_set


class MultiOutputDataset(Dataset):
    def __init__(self, filename):
        super().__init__(filename)
        self.tuple_ids = None
        self.tuple_id_to_tuple = None
        self.tuple_to_tuple_id = None
        self.unique_labels = None
        self.list_id_to_list = None
        self.zipped = None

    def compute_accuracy(self, y_pred):
        self.check_loaded()
        num_correct = 0
        for i in range(len(y_pred)):
            pred = y_pred[i]
            if pred is None:
                return None
            correct_tuples = set([tuple(l) for l in self.get_zipped()[i] if l[0] != -1])
            if set.issubset(make_set(pred), correct_tuples):
                num_correct += 1
        return num_correct / len(y_pred)

    '''
        array([[[ 0, -1, -1],
            [ 0, -1, -1],
            [ 0, -1, -1],
            [ 1,  2,  0]],
    
           [[ 0, -1, -1],
            [ 0, -1, -1],
            [ 0, -1, -1],
            [ 0,  0,  0]]])

        gets mapped to
        
        [[[ 0  0]
          [-1 -1]
          [-1 -1]]
        
         [[ 0  0]
          [-1 -1]
          [-1 -1]]
        
         [[ 0  0]
          [-1 -1]
          [-1 -1]]
        
         [[ 1  0]
          [ 2  0]
          [ 0  0]]]
    
        gets mapped to
        
        [[2 0 0]
         [2 0 0]
         [2 0 0]
         [3 4 2]]
    '''

    def get_zipped(self):
        if self.zipped is None:
            self.zipped = np.stack(self.Y_train, axis=2)
        return self.zipped

    def get_tuple_ids(self):
        if self.tuple_ids is not None:
            return self.tuple_ids

        stacked_y_train = self.get_zipped()

        # default
        tuple_to_index = {tuple(-1 for i in range(stacked_y_train.shape[2])): -1}

        self.tuple_ids = np.full((stacked_y_train.shape[0], stacked_y_train.shape[1]), -1)

        # first axis: datapoints
        # second axis: non-det
        # third axis: control inputs
        i = 0
        for datapoint in stacked_y_train:
            j = 0
            for action in datapoint:
                action_tuple = tuple(action)
                if action_tuple not in tuple_to_index.keys():
                    # indexing from 1
                    tuple_to_index[action_tuple] = len(tuple_to_index) + 1
                self.tuple_ids[i][j] = tuple_to_index[action_tuple]
                j = j + 1
            i = i + 1

        self.tuple_to_tuple_id = tuple_to_index
        self.tuple_id_to_tuple = {y: x for (x, y) in tuple_to_index.items()}
        return self.tuple_ids

    '''
    A list mapping tuple ids to list (float_id, float_id) tuples
    eg. {-1: (-1, -1), 2: (1, 3), 3: (2, 4)}
    '''

    def map_tuple_id_back(self, label):
        return self.tuple_id_to_tuple[label]

    def get_unique_labels(self):
        if self.unique_labels is None:
            self.unique_labels, self.list_id_to_list = self._get_unique_labels(self.get_tuple_ids())
        return self.unique_labels

    def map_unique_label_back(self, label):
        l = self.list_id_to_list[label]
        return [self.tuple_id_to_tuple[i] for i in l if i != -1]

    '''
    Generate a list of tuples (ctrl_idx, inp_enc, freq)
    where ctrl_idx is the control input index, inp_enc is the control input integer encoding and
    freq is the number of times the respective control input has occurred as the ctrl_idx'th component
    '''

    @staticmethod
    def _get_ranks(y):
        ranks = []
        for ctrl_idx in range(y.shape[0]):
            flattended_control = y[ctrl_idx].flatten()
            flattended_control = flattended_control[flattended_control != -1]
            counter = list(zip(range(len(np.bincount(flattended_control))), np.bincount(flattended_control)))
            idx_input_count = [(ctrl_idx,) + l for l in counter]
            ranks.extend(idx_input_count)
        return sorted(ranks, key=itemgetter(2), reverse=True)

    '''
    Given a y_train such as
    array([[[ 1,  2,  3],
            [ 1,  2,  1],
            [ 1,  2,  2],
            [ 3,  3, -1]],

           [[ 3,  4,  5],
            [ 3,  4,  4],
            [ 2,  6,  1],
            [ 3,  6, -1]]])

    gets determinized to

    array([[[ 1, -1, -1],
            [ 1, -1, -1],
            [ 1, -1, -1],
            [ 3, -1, -1]],

           [[ 3, -1, -1],
            [ 3, -1, -1],
            [ 2, -1, -1],
            [ 3, -1, -1]]])
            
    which is reduced to 
    
    array([[[1],
            [1],
            [1],
            [3]],
    
           [[3],
            [3],
            [2],
            [3]]])
    '''

    @staticmethod
    def determinize_max_over_all_inputs(y_original, tuple_to_tuple_id):
        y = np.copy(y_original)
        determinized = False

        # list of tuples (ctrl_input_idx, input_encoding) which were already considered for keeping
        already_considered = set()

        while not determinized:
            ranks = MultiOutputDataset._get_ranks(y)

            # find the ctrl_idx and inp_enc which should be used in the next round of pruning
            # i.e. the first one from the ranking list whose input has not been already considered
            ctrl_idx = None
            inp_enc = None
            for (ctr, inp, _) in ranks:
                if (ctr, inp) not in already_considered:
                    already_considered.add((ctr, inp))
                    ctrl_idx = ctr
                    inp_enc = inp
                    break

            # Go through y[ctrl_idx] row by row
            # for each row, if it contains input_encoding, then change the remaining into -1
            # make the same -1 changes for rest of the control inputs
            for i in range(y.shape[1]):
                row = y[ctrl_idx, i]
                if inp_enc in row:
                    for j in range(y.shape[2]):
                        if row[j] != inp_enc:
                            y[:, i, j] = -1

            # check if all rows contain only one element
            determinized = True
            for ctrl_idx in range(y.shape[0]):
                for i in range(y.shape[1]):
                    row = y[ctrl_idx, i]
                    valid_row = row[row != -1]
                    determinized = determinized & (valid_row.size == 1)
        valid_y = np.array([np.array([yyy[yyy != -1] for yyy in yy]) for yy in y])
        zipped = np.stack(valid_y, axis=2)
        # [[[1,3], [1,3], [1,2], [3,3]]] -> [1,1,2,3] (tuple ids)
        return np.apply_along_axis(lambda x: tuple_to_tuple_id[tuple(x)], axis=2, arr=zipped).flatten()



from dtcontrol.dataset.dataset import Dataset
from dtcontrol.util import make_set

class SingleOutputDataset(Dataset):
    def __init__(self, filename):
        super().__init__(filename)
        self.unique_labels = None
        self.unique_mapping = None

    def compute_accuracy(self, Y_pred):
        self.check_loaded()
        num_correct = 0
        for i in range(len(Y_pred)):
            pred = Y_pred[i]
            if pred is None:
                return None
            if set.issubset(make_set(pred), set(self.Y_train[i])):  # TODO: debug this once we have the tree in python. OC1 console says acc = 100%. i = 24455, pred = [3 4 0 0]
                num_correct += 1
        return num_correct / len(Y_pred)

    '''
    eg. 
    [[1  2  3 -1 -1],
     [1 -1 -1 -1 -1],
     [1  2 -1 -1 -1],
    ]
    
    gets mapped to
     
    unique_labels = [1, 2, 3]
    unique_mapping = {1: [1 2 3 -1 -1], 2: [1 -1 -1 -1 -1], 3: [1 2 -1 -1 -1]}
    TODO change unset to -1 (currently its 0)
    '''
    def get_unique_labels(self):
        if self.unique_labels is None:
            self.unique_labels, self.unique_mapping = self._get_unique_labels(self.Y_train)
        return self.unique_labels

    def map_unique_label_back(self, label):
        return self.unique_mapping[label]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dtcontrol",
    version="1.0.0rc4",
    description="A small tool which can convert automatically synthesised formally verified "
                "controllers into concise decision trees.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Mathias Jackermeier',
    author_email='mathias.jackermeier@outlook.de',
    license='MIT',
    url="https://gitlab.lrz.de/i7/dtcontrol",
    packages=['dtcontrol', 'dtcontrol.classifiers', 'dtcontrol.dataset', 'dtcontrol.ui',
              'dtcontrol.classifiers.OC1_source',
              'dtcontrol.c_templates'
              ],
    entry_points={
        'console_scripts': ['dtcontrol=dtcontrol.cli:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        'License :: OSI Approved :: MIT License'
    ],
    python_requires='>=3.6',
    install_requires=[
        'Jinja2==2.10.3',
        'pandas>=0.25.2',
        'pydot==1.4.1',
        'scikit-learn>=0.22',
        'tqdm>=4.36.1'
    ],
    package_data={
        'dtcontrol.c_templates': ['*.c'],
        'dtcontrol.ui': ['*.js', '*.css', '*.html', '*.py'],
        'dtcontrol.classifiers.OC1_source': ['*.c', '*.h', 'makefile', '*.readme', 'README'],
    }
)



import os
import sys
import subprocess
from collections import defaultdict
from shutil import copyfile

import numpy as np

import dtcontrol.util
from dtcontrol.classifiers.cart_custom_dt import CartNode
from dtcontrol.classifiers.custom_dt import CustomDT
from dtcontrol.classifiers.oc1_node import OC1Node
from dtcontrol.classifiers.oc1_parser import OC1Parser


class OC1Wrapper(CustomDT):
    """
    A wrapper for the OC1 C code.

    Make sure that you have compiled the code ('make mktree' in OC1_source)!
    """

    def __init__(self, num_restarts=40, num_jumps=20):
        super().__init__()
        self.name = 'OC1'
        self.oc1_path = 'classifiers/OC1_source/mktree'
        self.output_file = 'oc1_tmp/oc1_output'
        self.data_file = 'oc1_tmp/oc1_data.csv'
        self.dt_file = 'oc1_tmp/oc1_dt'
        self.log_file = 'oc1_tmp/oc1.log'
        self.num_restarts = num_restarts
        self.num_jumps = num_jumps
        self.num_nodes = None
        self.num_oblique = None
        self.current_dataset = None
        self.root: OC1Node = None
        self.oc1_parser = OC1Parser()
        self.oc1_reported_acc = None  # The accuracy reported by OC1. Used for debugging.
        self.num_extended_nodes = 0  # The amount of nodes added after OC1 to ensure full overfitting.

        if not os.path.exists(self.oc1_path):
            self.compile_oc1()

        if not os.path.exists('oc1_tmp'):
            os.mkdir('oc1_tmp')

    def compile_oc1(self):
        for path in dtcontrol.__path__:
            oc1_src = f"{path}/classifiers/OC1_source"
            if os.path.exists(oc1_src):
                if os.path.exists(oc1_src + "/mktree"):
                    self.oc1_path = oc1_src + "/mktree"
                    return
                try:
                    print(f"Compiling OC1...")
                    subprocess.call("make", cwd=oc1_src)
                    self.oc1_path = oc1_src + "/mktree"
                    print("Compiled OC1\n")
                except subprocess.CalledProcessError:
                    print("Compiling OC1 failed")  # todo use logging
                    sys.exit(-1)

    def is_applicable(self, dataset):
        return True

    def get_stats(self) -> dict:
        return {
            'nodes': self.num_nodes,
            'oblique': self.num_oblique,
            'extended': self.num_extended_nodes,
            'bandwidth': int(np.ceil(np.log2((self.num_nodes + 1) / 2)))
        }

    def get_fit_command(self, dataset):
        self.current_dataset = dataset
        self.save_data_to_file(np.c_[dataset.X_train, dataset.get_unique_labels()])
        command = '{} -t {} -D {} -p0 -i{} -j{} -l {}' \
            .format(self.oc1_path, self.data_file, self.dt_file, self.num_restarts, self.num_jumps, self.log_file)
        return command

    def fit_command_called(self):
        self.parse_fit_output()
        self.root: OC1Node = self.oc1_parser.parse_dt(self.dt_file)
        self.train_cart_nodes()
        self.root.set_labels(lambda x: self.current_dataset.map_unique_label_back(x.trained_label),
                             self.current_dataset.index_to_value)
        self.current_dataset = None

    def train_cart_nodes(self):
        node_to_data = defaultdict(list)
        node_to_labels = defaultdict(list)
        nodes_to_extend = []
        for i in range(len(self.current_dataset.X_train)):
            row = self.current_dataset.X_train[i]
            label = self.current_dataset.get_unique_labels()[i]
            parent = None
            node = self.root
            while node.left:
                parent = node
                node = node.left if node.test_condition(row.reshape(1, -1)) else node.right
            node_to_data[node].append(row)
            node_to_labels[node].append(label)
            if node.trained_label != label and node not in [n for n, _ in nodes_to_extend]:
                nodes_to_extend.append((node, parent))

        for node, parent in nodes_to_extend:
            cart_node = CartNode(node.depth)
            if parent.left == node:
                parent.left = cart_node
            else:
                parent.right = cart_node
            cart_node.fit(np.array(node_to_data[node]), np.array(node_to_labels[node]))
            self.num_nodes += cart_node.num_nodes - 1
            self.num_extended_nodes += cart_node.num_nodes - 1

    def fit(self, dataset):
        command = self.get_fit_command(dataset)
        self.execute_command(command)
        self.fit_command_called()

    def parse_fit_output(self):
        with open(self.output_file, 'r') as f:
            for line in f:
                if line.startswith('Number of nodes'):
                    self.num_nodes = int(line.split(': ')[1])
                elif line.startswith('Number of oblique'):
                    self.num_oblique = int(line.split(': ')[1])
                elif line.startswith('acc.'):
                    line = line.split('\t')[0]
                    line = line.split(' = ')[1]
                    self.oc1_reported_acc = float(line)

    def save_data_to_file(self, data):
        num_float_columns = data.shape[1] - 1
        np.savetxt(self.data_file, data, fmt=' '.join(['%f'] * num_float_columns + ['%d']), delimiter='\t')

    def execute_command(self, command):
        with open(self.output_file, 'w+') as out:
            p = subprocess.Popen(command.split(' '), stdout=out)
            p.wait()

    def save(self, filename):
        copyfile(self.dt_file, filename)

    def export_vhdl(self, numInputs, file=None):
        pass



from dtcontrol.dataset.multi_output_dataset import MultiOutputDataset
from dtcontrol.classifiers.cart_custom_dt import CartDT, CartNode


class MaxFreqMultiDT(CartDT):
    def __init__(self,):
        super().__init__()
        self.name = 'MaxFreqMultiDT'

    def is_applicable(self, dataset):
        return isinstance(dataset, MultiOutputDataset) and not dataset.is_deterministic

    def fit(self, dataset):
        if dataset.tuple_to_tuple_id is None:
            dataset.get_tuple_ids()
        self.root = MaxFreqMultiNode(dataset.tuple_to_tuple_id)
        self.root.fit(dataset.X_train, dataset.Y_train)
        self.set_labels(lambda leaf: dataset.map_tuple_id_back(leaf.trained_label), dataset.index_to_value)


class MaxFreqMultiNode(CartNode):
    def __init__(self, tuple_to_tuple_id, depth=0):
        super().__init__(depth)
        self.tuple_to_tuple_id = tuple_to_tuple_id

    def create_child_node(self):
        return MaxFreqMultiNode(self.tuple_to_tuple_id, self.depth + 1)

    def find_split(self, X, y):
        return super().find_split(X, MultiOutputDataset.determinize_max_over_all_inputs(y, self.tuple_to_tuple_id))

    def check_done(self, X, y):
        return super().check_done(X, MultiOutputDataset.determinize_max_over_all_inputs(y, self.tuple_to_tuple_id))



import re
import logging

import numpy as np
import pandas as pd

from dtcontrol.dataset.dataset_loader import DatasetLoader
from tqdm import tqdm


class UppaalDatasetLoader(DatasetLoader):
    def _load_dataset(self, filename):
        """
            Some assumptions regarding the Uppaal dataset
            1. All controllable states are named *.Choose
            2. Uppaal only works on integer domains (specifically np.int16)
            3. Actions are single-dimensional. However, if multiple components
               may have the .Choose state, they are added as categorical variables
               in X_train
            4. Step sizes for all state variables is 1
        """

        f = open(filename)
        print("Reading from %s" % filename)

        action_set = set()
        controllable_state_set = set()

        num_lines = sum(1 for line in f)

        f.seek(0)

        logging.info("Extracting actions and controllable components from UPPPAL dump")
        for i, line in enumerate(tqdm(f, total=num_lines)):
            if i == 1:
                # Extract numeric features
                numeric_features = re.findall(r'(\w+)=-?[0-9]+', line)

            # Extract actions and controllable components
            if line.startswith('State: '):
                ctrl = [word for word in line.split() if "Choose" in word]
                if ctrl:
                    controllable_state_set.update(ctrl)
            if line.startswith('When'):
                action_set.add(line[line.index(' take transition ') + 17:].rstrip())

        action_set = {action for action in action_set if "Choose" in action}
        actions = dict(zip(list(action_set), range(1, len(action_set) + 1)))
        controllable_states = list(controllable_state_set)

        # Figure out the assignments in each action and extract
        # the assigned value. The assigned variable is extracted
        # too, but not used anywhere as of now.
        index_to_value = dict()
        action_var = None
        for (action, index) in actions.items():
            _, action_var, val = re.findall(r"((\w+) := (-?[0-9]+))", action)[0]
            index_to_value[index] = int(val)

        row_num_vals = []
        row_actions = []

        ignore_current = False
        current_actions = []
        total_rows = 0
        total_state_actions = 0

        f.seek(0)

        for i in range(7):
            f.readline()

        logging.info("Extracting state-action pairs from UPPAAL dump")
        for i, line in enumerate(tqdm(f, total=num_lines)):
            if line.startswith("State:"):
                # find if the state is controllable, and if so, then make that position 1 in categorical vals
                controllable = False
                cat_vals = [0 for i in range(len(controllable_states))]
                for i in range(len(controllable_states)):
                    if controllable_states[i] in line:
                        cat_vals[i] = 1
                        controllable = True
                if not controllable:
                    ignore_current = True
                    continue
                else:
                    ignore_current = False
                    numeric_vals = [word.split("=")[1] for word in line.split() if "=" in word]
            elif ignore_current:
                continue
            elif line.startswith("When"):
                action_str = line[line.index(' take transition ') + 17:].rstrip()
                current_actions.append(actions[action_str])
            elif line.startswith("While"):
                # We implicityly assume that transitions starting with 'While' are mapped to wait.
                ignore_current = True
            elif line.strip() == "":
                if not ignore_current:
                    row_num_vals.append(cat_vals + numeric_vals)
                    row_actions.append(current_actions)
                    total_rows += 1
                    total_state_actions += len(current_actions)
                ignore_current = False
                current_actions = []
            else:
                raise Exception("ERROR: Unhandled line in input")

        logging.info(f"Done reading {total_rows} states with \na total of {total_state_actions} state-action pairs.")

        # Project onto measurable variables, the strategy should not depend on the gua variables coming from euler
        projection_variables = controllable_states + list(filter(lambda x: 'gua' not in x, numeric_features))
        num_df = pd.DataFrame(row_num_vals, columns=controllable_states + numeric_features, dtype=np.int16)
        num_df = num_df[projection_variables]

        grouped = num_df.groupby(projection_variables)
        X = np.empty((len(grouped), len(projection_variables)), dtype=np.int16)
        Y = np.full((len(grouped), max([len(y) for y in row_actions])), -1, dtype=np.int16)

        i = 0
        for (group, indices) in grouped.indices.items():
            if len(indices) > 1:
                X[i] = group
                conservative_actions = set(actions.values()).copy()
                for idx in indices:
                    conservative_actions &= set(row_actions[idx])
                assert len(
                    conservative_actions) > 0, "Stategy for picking safe action doesn't work. Deeper analysis needed."
                Y[i][0:len(conservative_actions)] = list(sorted(conservative_actions))
            else:
                X[i] = group
                Y[i][0:len(row_actions[indices[0]])] = sorted(row_actions[indices[0]])
            i = i + 1

        print("Constructed training set with %s datapoints" % X.shape[0])

        # construct metadata
        # assumption is that UPPAAL only works with integers
        X_metadata = dict()
        X_metadata["variables"] = projection_variables
        X_metadata["min"] = [int(i) for i in np.amin(X, axis=0)]
        X_metadata["max"] = [int(i) for i in np.amax(X, axis=0)]
        X_metadata["step_size"] = [1 for _ in range(len(projection_variables))]

        Y_metadata = dict()
        Y_metadata["variables"] = [action_var]
        Y_metadata["min"] = [min(index_to_value.values())]
        Y_metadata["max"] = [max(index_to_value.values())]
        Y_metadata["step_size"] = [int((Y_metadata["max"][0] - Y_metadata["min"][0])/(len(index_to_value) - 1))]

        logging.debug(X_metadata)
        logging.debug(Y_metadata)

        return (X, X_metadata, Y, Y_metadata, index_to_value)



import warnings

import numpy as np
from sklearn.exceptions import ConvergenceWarning
from sklearn.tree import DecisionTreeClassifier

import dtcontrol.util
from dtcontrol.classifiers.custom_dt import CustomDT, Node

# SVM sometimes does not converge and clutters the output with warnings
warnings.filterwarnings("ignore", category=ConvergenceWarning)

class LinearClassifierDT(CustomDT):
    def __init__(self, classifier_class, **kwargs):
        super().__init__()
        self.kwargs = kwargs
        self.classifier_class = classifier_class
        self.name = 'LinearClassifierDT-{}'.format(classifier_class.__name__)

    def is_applicable(self, dataset):
        return True

    def fit(self, dataset):
        self.root = LinearClassifierOrAxisAlignedNode(self.classifier_class, **self.kwargs)
        self.root.fit(dataset.X_train, dataset.get_unique_labels())
        self.set_labels(lambda leaf: dataset.map_unique_label_back(leaf.trained_label), dataset.index_to_value)

    def get_stats(self) -> dict:
        return {
            'nodes': self.root.num_nodes,
            'oblique': self.root.num_not_axis_aligned,
            'bandwidth': int(np.ceil(np.log2((self.root.num_nodes + 1)/2)))
        }

class LinearClassifierOrAxisAlignedNode(Node):
    def __init__(self, classifier_class, depth=0, **kwargs):
        super().__init__(depth)
        self.classifier_class = classifier_class
        self.classifier = None
        self.kwargs = kwargs
        self.num_not_axis_aligned = 0

    def test_condition(self, x):
        if not self.is_axis_aligned():
            return self.classifier.predict(x)[0] == -1
        else:
            tree = self.classifier.tree_
            return x[:, tree.feature[0]][0] <= tree.threshold[0]

    def create_child_node(self):
        return LinearClassifierOrAxisAlignedNode(self.classifier_class, self.depth + 1, **self.kwargs)

    def fit_children(self, X, y, mask):
        super().fit_children(X, y, mask)
        self.num_not_axis_aligned = self.num_not_axis_aligned + self.left.num_not_axis_aligned + \
                                    self.right.num_not_axis_aligned

    def find_split(self, X, y):
        lc, lc_impurity, lc_mask = self.find_best_linear_classifier(X, y)
        dt, dt_impurity, dt_mask = self.find_best_axis_aligned_split(X, y)
        if dt_impurity <= lc_impurity:
            self.classifier = dt
            mask = dt_mask
        else:
            self.num_not_axis_aligned += 1
            self.classifier = lc
            mask = lc_mask
        return mask

    def find_best_linear_classifier(self, X, y):
        label_to_impurity = {}
        label_to_classifier = {}
        for label in np.unique(y):
            new_y = np.copy(y)
            label_mask = (new_y == label)
            new_y[label_mask] = 1
            new_y[~label_mask] = -1
            classifier = self.classifier_class(**self.kwargs)
            classifier.fit(X, new_y)
            label_to_classifier[label] = classifier
            pred = classifier.predict(X)
            impurity = self.calculate_impurity(y, (pred == -1))
            label_to_impurity[label] = impurity

        min_impurity = min(label_to_impurity.values())
        label = min(label_to_impurity.items(), key=lambda x: x[1])[0]
        classifier = label_to_classifier[label]
        mask = (classifier.predict(X) == -1)
        return classifier, min_impurity, mask

    def find_best_axis_aligned_split(self, X, y):
        dt = DecisionTreeClassifier(max_depth=1, criterion='entropy')
        dt.fit(X, y)
        dt_mask = X[:, dt.tree_.feature[0]] <= dt.tree_.threshold[0]
        return dt, self.calculate_impurity(y, dt_mask), dt_mask

    def is_axis_aligned(self):
        return not isinstance(self.classifier, self.classifier_class)

    def get_dot_label(self):
        if self.actual_label is not None:
            rounded = dtcontrol.util.objround(self.actual_label, 6) # TODO get precision from flag?
            return dtcontrol.util.split_into_lines(rounded)
        if self.is_axis_aligned():
            tree = self.classifier.tree_
            return f'X[{tree.feature[0]}] <= {round(tree.threshold[0], 6)}'  # TODO get precision from flag?
        else:
            # this implicitly assumes n_classes == 2
            coef_ = self.classifier.coef_[0]
            intercept_ = self.classifier.intercept_[0]
            line = []
            for i in range(len(coef_)):
                line.append(f"{round(coef_[i], 6)}*X[{i}]")  # TODO get precision from flag?
            line.append(f"{round(intercept_, 6)}")  # TODO get precision from flag?
            hyperplane = "\n+".join(line) + " <= 0"
            return hyperplane.replace('+-', '-')

    def print_dot_red(self):
        return False



import numpy as np

from dtcontrol.classifiers.cart_custom_dt import CartDT, CartNode
from dtcontrol.dataset.single_output_dataset import SingleOutputDataset

class RandomDT(CartDT):

    def __init__(self):
        super().__init__()
        self.name = 'RandomDT'

    def is_applicable(self, dataset):
        return not dataset.is_deterministic

    def fit(self, dataset):
        self.root = CartNode()
        if isinstance(dataset, SingleOutputDataset):
            self.root.fit(dataset.X_train, self.determinize(dataset.Y_train))
            self.set_labels(lambda leaf: leaf.trained_label, dataset.index_to_value)
        else:
            self.root.fit(dataset.X_train, self.determinize(dataset.get_tuple_ids()))
            self.set_labels(lambda leaf: dataset.map_tuple_id_back(leaf.trained_label), dataset.index_to_value)

    def determinize(self, y):
        return np.apply_along_axis(lambda x: np.random.choice(x[x != -1]), axis=1, arr=y)



import ast
import re

from dtcontrol.classifiers.cart_custom_dt import CartNode
from dtcontrol.classifiers.oc1_node import OC1Node
from dtcontrol.util import peek_line

class OC1Parser:
    def parse_dt(self, dt_file):
        """
        Parses the OC1 generated DT file
        :return OC1Node object pointing to the root
        """
        with open(dt_file) as infile:
            dim = int(re.findall(r"Dimensions: ([0-9]+)", infile.readline())[0])
            return self.parse_recursive(dim, infile, -1)

    def parse_recursive(self, dim, infile, prev_depth):
        if not self.advance_to_next_hyperplane(infile):  # reached EOF
            node = OC1Node(None, None, prev_depth + 1)
            return node
        line = peek_line(infile)
        assert 'Hyperplane' in line
        path, left, right = self.parse_node(line)
        if prev_depth >= 0 and prev_depth >= len(path):
            node = OC1Node(None, None, prev_depth + 1)
            return node
        infile.readline()
        expr_line = infile.readline()
        coeff, intercept = self.parse_expression(expr_line, dim)
        depth = prev_depth + 1
        node = OC1Node(coeff, intercept, depth)
        if self.is_pure(left) and self.is_pure(right):
            left_node = OC1Node(None, None, depth + 1)
            left_node.trained_label = self.get_leaf_class(left)
            left_node.path = path + 'l'
            right_node = OC1Node(None, None, depth + 1)
            right_node.trained_label = self.get_leaf_class(right)
            right_node.path = path + 'r'
            node.left = left_node
            node.right = right_node
        else:
            if self.is_pure(left):
                left_node = OC1Node(None, None, depth + 1)
                left_node.trained_label = self.get_leaf_class(left)
                node.left = left_node
                left_node.path = path + 'l'
                node.right = self.parse_recursive(dim, infile, depth)
            elif self.is_pure(right):
                right_node = OC1Node(None, None, depth + 1)
                right_node.trained_label = self.get_leaf_class(right)
                node.right = right_node
                right_node.path = path + 'r'
                node.left = self.parse_recursive(dim, infile, depth)
            else:
                node.left = self.parse_recursive(dim, infile, depth)
                node.right = self.parse_recursive(dim, infile, depth)
        return node

    @staticmethod
    def advance_to_next_hyperplane(infile):
        while 'Hyperplane' not in peek_line(infile):
            pos = infile.tell()
            infile.readline()
            if infile.tell() == pos:  # EOF
                return False
        return True

    @staticmethod
    def parse_expression(line, dim):
        # remove the " = 0" part and the split
        summands = line[:-5].split(' + ')

        intercept = float(summands[-1])

        j = 0
        coeffs = []
        for i in range(1, dim + 1):
            if f"x[{i}]" in summands[j]:
                coeffs.append(float(summands[j].split()[0]))
                j = j + 1
            else:
                coeffs.append(0)
        return coeffs, intercept

    @staticmethod
    def is_pure(list_):
        return len(list_) - list_.count(0) == 1

    @staticmethod
    def get_leaf_class(leaf):
        for i in range(len(leaf)):
            if leaf[i] != 0:
                return i + 1

    @staticmethod
    def parse_node(line):
        decision_path, _, _, _, left, _, _, right = line.split()
        return decision_path, ast.literal_eval(left[:-1]), ast.literal_eval(right)



import multiprocessing
import os
import signal
import subprocess
import time


def is_windows():
    return os.name == 'nt'


def use_multiprocessing():
    return is_windows()


def _call_and_return_in_list(obj, method, return_list, *args):
    getattr(obj, method)(*args)
    return_list[0] = obj


def call_with_timeout(obj, method, *args, timeout=60):
    """
    Calls a method on an object and stops the execution once the time limit is reached.
    :param obj: The object on which the method is to be called
    :param method: The method to be called on the object
    :param timeout: The time limit
    :param args: The arguments to be passed to the method
    :return: The modified object, True if the call was successful and False otherwise, and the actual time needed
    """

    # this slight hack is used to properly terminate the OC1Wrapper, which needs to spawn a child process itself
    if hasattr(obj, 'get_{}_command'.format(method)):
        command = getattr(obj, 'get_{}_command'.format(method))(*args)
        with open(obj.output_file, 'w+') as out:
            start = time.time()
            p = subprocess.Popen(command.split(' '), stdout=out, stderr=out)
            try:
                p.wait(timeout=timeout)
            except subprocess.TimeoutExpired:
                p.kill()
                return obj, False, time.time() - start
        getattr(obj, '{}_command_called'.format(method))()
        return obj, True, time.time() - start
    else:
        if use_multiprocessing():
            with multiprocessing.Manager() as manager:
                return_list = manager.list(range(1))
                p = multiprocessing.Process(target=_call_and_return_in_list, args=(obj, method, return_list, *args))
                start = time.time()
                p.start()
                p.join(timeout)
                if p.is_alive():
                    p.terminate()
                    p.join()
                    return obj, False, time.time() - start
                return return_list[0], True, time.time() - start
        else:
            def raise_timeout(signum, frame):
                raise TimeoutError

            signal.signal(signal.SIGALRM, raise_timeout)
            signal.alarm(timeout)
            start = time.time()
            try:
                getattr(obj, method)(*args)
                return obj, True, time.time() - start
            except TimeoutError:
                return obj, False, time.time() - start
            finally:
                signal.signal(signal.SIGALRM, signal.SIG_IGN)
